// wie include/Datum.h, aber mit neuen op<<, op>>, toString()
#ifndef DATUM_H
#define DATUM_H
#include <ctime>
#include <iostream>
#include <locale>
#include <string>

class UngueltigesDatumException : public std::runtime_error {
public:
  UngueltigesDatumException(int t, int m, int j)
      : std::runtime_error(toString(t, m, j))
  {
  }

private:
  static std::string toString(int tag, int monat, int jahr)
  {
    std::string meldung = std::to_string(tag) + "." + std::to_string(monat) + "." +
                          std::to_string(jahr) + " ist ein ungueltiges Datum!";
    return meldung;
  }
};

constexpr bool istSchaltjahr(int jahr)
{
  return (jahr % 4 == 0 && jahr % 100) || jahr % 400 == 0;
}

constexpr bool istGueltigesDatum(int tag, int monat, int jahr)
{
  bool ergebnis = false;
  if ((monat >= 1 && monat <= 12) &&
      (jahr > 1582              // 1582: Beginn des  gregorianischen Kalenders
       && jahr < 2400)          // Plausibilitätskontrolle
  ) {
    int letzterTagImMonat = 29;                // nur Februar im Schaltjahr
    if (monat != 2 || !istSchaltjahr(jahr)) {          // sonst
      constexpr int tageImMonat[]{31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
      letzterTagImMonat = tageImMonat[monat - 1];
    }
    ergebnis = (tag >= 1 && tag <= letzterTagImMonat);
  }
  return ergebnis;
}

class Datum {
public:
  Datum()
  {          // Standardkonstruktor
    aktuell();
  }
  constexpr Datum(int t, int m,
                  int j)          // allgemeiner Konstruktor
      : tag_(t), monat_(m), jahr_(j)
  {
    if (!istGueltigesDatum(t, m, j)) {
      throw UngueltigesDatumException(t, m, j);
    }
  }
  constexpr auto tag() const { return tag_; }
  constexpr auto monat() const { return monat_; }
  constexpr auto jahr() const { return jahr_; }
  constexpr bool istSchaltjahr() const
  {
    return ::istSchaltjahr(jahr_);          // globale Funktion mit ::-Operator
  }

  void set(int t, int m, int j)
  {
    if (!istGueltigesDatum(t, m, j)) {
      throw UngueltigesDatumException(t, m, j);
    }
    tag_ = t;
    monat_ = m;
    jahr_ = j;
  }

  void aktuell()
  {          // Systemdatum setzen
    // time_t, time(), tm, localtime() sind in <ctime> deklariert
    std::time_t now{std::time(nullptr)};     // aktuelle Zeit in s seit 1.1.1970
    std::tm* z{std::localtime(&now)};        // Zeiger auf eine vordefinierte Struktur des Typs tm.
    jahr_ = z->tm_year + 1900;
    monat_ = z->tm_mon + 1;                  // localtime liefert 0..11
    tag_ = z->tm_mday;
  }

  Datum& operator++()
  {          // Tag hochzählen, präfix
    ++tag_;
    if (!istGueltigesDatum(tag_, monat_,
                           jahr_)) {          // Monatsende erreicht?
      tag_ = 1;
      if (++monat_ > 12) {                    // Jahresende erreicht?
        monat_ = 1;
        ++jahr_;
      }
    }
    return *this;
  }

  Datum operator++(int)
  {          // Tag hochzählen, postfix
    Datum temp = *this;
    ++*this;                                 // Präfix ++ aufrufen
    return temp;
  }

  // alternativ: durch Methode string toString() ersetzen (siehe
  // Übungsaufgabe)
  explicit operator std::string() const
  {          // Typumwandlungsoperator
    std::string temp("tt.mm.jjjj");
    // implizite Umwandlung in char
    temp[0] = tag_ / 10 + '0';
    temp[1] = tag_ % 10 + '0';
    temp[3] = monat_ / 10 + '0';
    temp[4] = monat_ % 10 + '0';
    int pos = 9;          // letzte Jahresziffer
    int j = jahr_;
    while (j > 0) {
      temp[pos] = j % 10 + '0';          // letzte Ziffer
      j = j / 10;                        // letzte Ziffer abtrennen
      --pos;
    }
    return temp;
  }
  // std::string toString() const; neu:
  std::string toString(const std::locale& loc = std::locale("de_DE")) const;

private:
  int tag_;
  int monat_;
  int jahr_;
};

// neu
std::ostream& operator<<(std::ostream& os, const Datum& d);
std::istream& operator>>(std::istream&, Datum&);

constexpr bool operator==(const Datum& a, const Datum& b)
{
  return a.tag() == b.tag() && a.monat() == b.monat() && a.jahr() == b.jahr();
}

constexpr bool operator<(const Datum& a, const Datum& b)
{
  return a.jahr() < b.jahr() || (a.jahr() == b.jahr() && a.monat() < b.monat()) ||
         (a.jahr() == b.jahr() && a.monat() == b.monat() && a.tag() < b.tag());
}

#endif

